#! /usr/bin/env python3
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
import os.path
import yaml
import subprocess
import argparse

parser = argparse.ArgumentParser(
    description='Generate Glean telemetry definitions from YAML sources')
parser.add_argument('-o', '--output', metavar='DIR', type=str, action='store',
    default=os.path.join(os.getcwd(), 'glean', 'telemetry'),
    help='Output directory for generated files')
parser.add_argument('-j', '--javadir', metavar='DIR', type=str, action='store',
    help='Output directory for Java/Kotlin files')
args = parser.parse_args()


def camelize(string):
    output = ''
    first = True
    for chunk in string.split('_'):
        if first:
            output += chunk
            first = False
        else:
            output += chunk[0].upper()
            output += chunk[1:]
    return output


## Find the input Yaml files relative to the script location.
rootpath = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, os.pardir))
metrics = os.path.join(rootpath, 'glean', 'metrics.yaml')
pings = os.path.join(rootpath, 'glean', 'pings.yaml')

yaml_file = open(metrics, 'r', encoding='utf-8')
yaml_content = yaml.safe_load(yaml_file)


print("Parsing the metrics.yaml file...")
if not os.path.isfile(metrics):
    print(f'Unable to find {metrics}')
    exit(1)

## Create the output directory if it doesn't already exist.
os.makedirs(args.output, exist_ok=True)


print("Generating the C++ header...")
output = open(os.path.join(args.output, "gleansample.h"), "w")
output.write("""/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

// AUTOGENERATED! DO NOT EDIT!!

namespace GleanSample {

""")
for key in yaml_content['sample']:
   sampleName = camelize(key)
   output.write(f"constexpr const char* {sampleName} = \"{sampleName}\";\n")

output.write("""
} // GleanSample
""");
output.close();


print("Generating the QRC file...")
output = open(os.path.join(args.output, "telemetry.qrc"), "w")
output.write("""<!-- AUTOGENERATED! DO NOT EDIT -->
<RCC>
    <qresource prefix="/glean/telemetry">
        <file>pings.js</file>
        <file>sample.js</file>
        <file>qmldir</file>
    </qresource>
</RCC>
""")
output.close()

if args.javadir is not None:
    print("Generating the Kotlin enum...")
    androiddir = os.path.join(args.javadir, 'org', 'mozilla', 'firefox', 'vpn', 'glean')
    os.makedirs(androiddir, exist_ok=True)
    output = open(os.path.join(androiddir, "GleanEvents.kt"), "w")
    output.write("""/* This Source Code Form is subject to the terms of the Mozilla Public
* License, v. 2.0. If a copy of the MPL was not distributed with this
* file, You can obtain one at http://mozilla.org/MPL/2.0/. */

// AUTOGENERATED! DO NOT EDIT!!
package org.mozilla.firefox.vpn.glean

import android.annotation.SuppressLint

@SuppressLint("Unused")
enum class GleanEvent(val key: String) {
""")
    for key in yaml_content['sample']:
        sampleName = camelize(key)
        output.write(f"\t {sampleName}(\"{sampleName}\"),\n")

    output.write("""
    } // GleanSample
    """)
    output.close()

print("Generating the JS modules...")
try:
    subprocess.call([sys.executable, "-m", "glean_parser", "translate", metrics, pings,
                     "-f", "javascript", "-o", args.output, "--option", "platform=qt",
                     "--option", "version=0.30"])
except Exception as e:
    print("glean_parser failed. Is it installed? Try with:\n\tpip3 install -r requirements.txt --user")
    print(e)
    exit(1)

